#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  Commands and their actions.
"""

import asyncio
import logging

class Commands:
    log = logging.getLogger('ad1459.commands')
    def me(self, room, client, message):
        """ /me command

        Sends a CTCP Action.

        Arguments:
            room (:obj:`Room`): The room to ACTION
            client (:obj:`Client`): The client object
            message (str): The action to send.
        """
        amessage = ' '.join(message.split()[1:])
        self.log.debug('Sending action to %s: %s', room.name, amessage)
        asyncio.run_coroutine_threadsafe(
            client.ctcp(room.name, 'ACTION', contents=amessage),
            loop=asyncio.get_event_loop()
        )



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This is the About Dialog.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from ad1459 import __version__, __license__


class AboutDialog(Gtk.AboutDialog):

    def __init__(self):
        super().__init__()

        self.set_program_name('AD1459')
        self.set_logo_icon_name('in.donotspellitgav.Ad1459')
        self.set_version(__version__.__version__)
        self.set_website('https://github.com/g4vr0che/ad1459')
        self.set_website_label('github.com/g4vr0che/ad1459')
        self.set_copyright('Copyright ©2019 Gaven Royer et al')
        self.set_license(__license__.__license__)
        self.set_wrap_license(True)
        self.set_authors(
            [
                'Gaven Royer'
            ]
        )



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This file is an application window.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Gdk': '3.0'
    }
)
from gi.repository import Gtk, Gdk

from .about import AboutDialog
from .headerbar import Headerbar
from .irc_entry import IrcEntry
from .room_switcher import RoomSwitcher
from .server_popup import ServerPopover

class Ad1459Window(Gtk.Window):
    """ This is a window for AD1459. It contains the overall layout as well as
    most of the general controls for the app.
    """

    def __init__(self, app, parser):
        self.log = logging.getLogger('ad1459.window')
        self.log.debug('Creating window')

        self.app = app
        self.parser = parser

        super().__init__()
        self.set_default_size(1000, 600)

        self.header = Headerbar(self.app)
        self.set_titlebar(self.header)

        self.main_pane = Gtk.HPaned()
        self.main_pane.set_position(200)
        self.add(self.main_pane)

        channel_grid = Gtk.Grid()
        channel_grid.set_hexpand(True)
        channel_grid.set_vexpand(True)
        self.main_pane.add2(channel_grid)

        switcher_grid = Gtk.Grid()
        switcher_grid.set_hexpand(True)
        switcher_grid.set_vexpand(True)
        self.main_pane.add1(switcher_grid)

        self.switcher = RoomSwitcher(self.app)
        switcher_grid.attach(self.switcher, 0, 0, 1, 1)

        join_bar = Gtk.ActionBar()
        switcher_grid.attach(join_bar, 0, 1, 1, 1)

        self.join_entry = Gtk.Entry()
        self.join_entry.set_width_chars(1)
        self.join_entry.set_hexpand(True)
        self.join_entry.set_placeholder_text('Join channel')
        self.join_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, 'list-add-symbolic'
        )
        self.join_entry.set_icon_activatable(
            Gtk.EntryIconPosition.SECONDARY, True
        )
        join_bar.pack_start(self.join_entry)

        entry_grid = Gtk.ActionBar()
        channel_grid.attach(entry_grid, 0, 1, 1, 1)
        
        self.irc_entry = IrcEntry(self)
        self.nick_button = Gtk.Button.new_with_label('nickname')
        self.send_button = Gtk.Button.new_from_icon_name(
            'mail-send-symbolic', Gtk.IconSize.BUTTON
        )

        self.nick_button.set_halign(Gtk.Align.START)
        self.send_button.set_halign(Gtk.Align.END)

        Gtk.StyleContext.add_class(
            self.nick_button.get_style_context(), 'flat'
        )
        Gtk.StyleContext.add_class(
            self.send_button.get_style_context(), 'suggested-action'
        )

        entry_grid.pack_start(self.nick_button)
        entry_grid.pack_start(self.irc_entry)
        entry_grid.pack_start(self.send_button)

        self.channel_pane = Gtk.HPaned()
        self.channel_pane.set_position(600)
        channel_grid.attach(self.channel_pane, 0, 0, 1, 1)

        self.message_stack = Gtk.Stack()
        self.message_stack.set_hexpand(True)
        self.message_stack.set_vexpand(True)
        self.message_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.message_stack.set_transition_duration(100)
        self.channel_pane.pack1(self.message_stack, True, True)

        self.topic_stack = Gtk.Stack()
        self.topic_stack.set_hexpand(True)
        self.topic_stack.set_vexpand(True)
        self.topic_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.topic_stack.set_transition_duration(100)
        self.channel_pane.pack2(self.topic_stack, False, False)

        self.irc_entry.grab_focus_without_selecting()

        self.connect('focus-in-event', self.on_focus_changed)
    
    # Internal Handlers
    def on_focus_changed(self, window, data=None):
        try:
            room = self.message_stack.get_visible_child().room
        except AttributeError:
            return
        self.log.debug('Unsetting unread indicator for %s', room.name)
        room.row.set_icon('radio-symbolic')

    # Data
    @property
    def active_room(self):
        """ Room: The currently active room."""
        room = self.message_stack.get_visible_child().room
        return room

    @property
    def focused(self):
        """bool: Whether or not the window has focus."""
        if self.props.is_active:
            return True
        
        return False



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  The topic pane contents.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0'
    }
)
from gi.repository import Gtk, GLib

from .user_row import UserRow

class TopicPane(Gtk.Grid):
    """ The contents of the topic pane, as a GtkGrid()."""

    def __init__(self, room):
        super().__init__()
        self.room = room
        self.log = logging.getLogger('ad1459.topic')
        self.topic_expander = Gtk.Expander()
        self.topic_expander.set_label_fill(True)
        self.attach(self.topic_expander, 0, 0, 1, 1)

        # Set Expander Label
        self.exp_label = Gtk.Label()
        self.exp_label.set_line_wrap(True)
        self.exp_label.set_xalign(0)
        self.exp_label.set_margin_start(3)
        self.exp_label.set_margin_end(3)
        self.topic_expander.set_label_widget(self.exp_label)

        self.topic_label = Gtk.Label()
        self.topic_label.set_margin_start(6)
        self.topic_label.set_margin_end(6)
        self.topic_label.set_margin_top(3)
        self.topic_label.set_margin_bottom(6)
        self.topic_label.set_xalign(0)
        self.topic_label.set_line_wrap(True)
        self.topic_expander.add(self.topic_label)

        separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        separator.set_hexpand(True)
        self.attach(separator, 0, 1, 1, 1)

        user_window = Gtk.ScrolledWindow()
        user_window.set_hexpand(True)
        user_window.set_vexpand(True)
        self.attach(user_window, 0, 2, 1, 1)

        self.user_list = Gtk.ListBox()
        self.user_list.set_hexpand(True)
        self.user_list.set_vexpand(True)
        self.user_list.set_sort_func(sort_users)
        user_window.add(self.user_list)

        self.update_topic()
    
    def update_topic(self):
        """Update the channel topic."""
        self.log.debug('Updating topic for %s', self.room.name)

        try:
            topic_expander_label = '<small>Set by '
            topic_by = GLib.markup_escape_text(self.room.data['topic_by'])
            topic_expander_label += f'<i>{topic_by}</i> on '
            topic_expander_label += f'<i>{self.room.data["topic_set"]}</i>:</small>'
            topic_text = GLib.markup_escape_text(self.room.data['topic'])
            topic_text = self.room.window.parser.parse_text(topic_text)
            topic_text = self.room.window.parser.hyperlinks(topic_text)
            self.topic_expander.set_sensitive(True)
            self.topic_expander.set_expanded(True)
        
        except (KeyError, TypeError):
            topic_expander_label = "No topic set"
            topic_text = ''
            self.topic_expander.set_sensitive(False)
            self.topic_expander.set_expanded(False)

        self.exp_label.set_markup(topic_expander_label)
        self.topic_label.set_markup(f'<small>{topic_text}</small>')

    def update_users(self):
        """Updates this room's user list."""
        curr_users = self.user_list.get_children()
        
        for user in curr_users:
            GLib.idle_add(user.destroy)

        for user in self.room.users:
            new_user = UserRow(self.room)
            new_user.nick = user
            self.user_list.add(new_user)
        
        self.room.window.show_all()
        self.user_list.invalidate_sort()

def sort_users(row1, row2, *user_data):
    if row1.nick.upper() < row2.nick.upper():
        return -1
    else:
        return 1


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  The IRC client object.
"""

import asyncio
import logging
import pydle
import time

class Client(pydle.Client):

    def __init__(self, nick, network, sasl_username=None, sasl_password=None, **kwargs):
        self.log = logging.getLogger('ad1459.client')
        super().__init__(nick, sasl_username=sasl_username, sasl_password=sasl_password, **kwargs)
        self.network_ = network
        self.log.debug('Created client for network %s', self.network_.name)
        self.bouncer = False
    
    async def connect(self, hostname=None, password=None, **kwargs):
        self.log.debug('Client initiating connection to %s', hostname)
        await super().connect(hostname=hostname, password=password, **kwargs)
    
    async def on_connect(self):
        self.log.info('Connected to %s', self.network_.name)
        await super().on_connect()
        await self.network_.on_connected()
    
    async def _disconnect(self, expected):
        """ We need to override this to stop the event loop getting closed."""
        # Shutdown connection.
        await self.connection.disconnect()

        # Reset any attributes.
        self._reset_attributes()

        # Callback.
        await self.on_disconnect(expected)

        """This is our override.
        # Shut down event loop.
        # if expected and self.own_eventloop:
        #     self.connection.stop()
        """
    
    async def on_raw(self, message):
        if message.command == ('CAP' or 'cap' or 'Cap'):
            if 'znc.in/' in " ".join(message.params):
                self.log.debug('Server appears to be a ZNC Bouncer')
                self.bouncer = True
        
        await super().on_raw(message)
    
    async def on_nick_change(self, old, new):
        self.log.debug('User %s is now %s', old, new)
        await self.network_.on_nick_change(old, new)
        await super().on_nick_change(old, new)
    
    async def on_join(self, channel, user):
        self.log.debug(f'User {user} joined {channel} on {self.network_.name}')
        await self.network_.on_join(channel, user)
        await super().on_join(channel, user)
    
    async def on_part(self, channel, user, message=None):
        self.log.debug(f'User {user} parted {channel} on {self.network_.name}')
        await self.network_.on_part(channel, user, message=message)
        await super().on_part(channel, user, message=message)
    
    async def on_kick(self, channel, target, by, reason=None):
        self.log.debug('%s was kicked from %s by %s (%s)', target, channel, by, reason)
        await self.network_.on_kick(channel, target, by, reason=reason)
        await super().on_kick(channel, target, by, reason=reason)
    
    async def on_quit(self, user, message=None):
        self.log.debug(f'User {user} has quit {self.network_.name}')
        await self.network_.on_quit(user, message=message)
        await super().on_quit(user, message=message)
    
    async def on_kill(self, target, by, reason):
        self.log.debug('%s was killed by %s, (%s)', target, by, reason)
        await self.network_.on_kill(target, by, reason=reason)
        super().on_kill(target, by, reason)

    async def on_channel_message(self, target, source, message):
        self.log.debug('New message in %s from %s: %s', target, source, message)
        await self.network_.on_message(target, source, message)
        await super().on_message(target, source, message)
    
    async def on_channel_notice(self, target, by, message):
        self.log.debug('Received notice to %s from %s', target, by)
        await self.network_.on_notice(target, by, message)
        await super().on_notice(target, by, message)
    
    async def on_private_message(self, target, by, message):
        self.log.debug('New private message to %s from %s: %s', target, by, message)
        await self.network_.on_private_message(target, by, message)
        await super().on_private_message(target, by, message)
    
    async def on_private_notice(self, target, by, message):
        self.log.debug('Received private notice to %s from %s', target, by)
        await self.network_.on_private_notice(target, by, message)
        await super().on_private_notice(target, by, message)
    
    async def on_mode_change(self, channel, modes, by):
        self.log.debug('%s set modes %s in %s', by, modes, channel)
        await self.network_.on_mode_change(channel, modes, by)
        await super().on_mode_change(channel, modes, by)
    
    async def on_topic_change(self, channel, message, by):
        self.log.debug('Topic in %s set to "%s" by %s', channel, message, by)
        await self.network_.on_topic_change(channel, message, by)
        await super().on_topic_change(channel, message, by)
    
    async def on_user_invite(self, target, channel, by):
        self.log.debug('%s was invited to %s by %s', target, channel, by)
        await self.network_.on_user_invite(target, channel, by)
        super().on_user_invite(target, channel, by)

    async def on_invite(self, channel, by):
        self.log.debug('Invited to %s by %s', channel, by)
        await self.network_.on_invite(channel, by)
        await super().on_invite(channel, by)
    
    # CTCP Support

    async def on_ctcp(self, by, target, what, contents):
        self.log.debug('%s received CTCP %s from %s: %s', target, what, by, contents)
        super().on_ctcp(by, target, what, contents)
    
    async def on_ctcp_reply(self, by, target, what, contents):
        self.log.debug('%s received REPLY CTCP %s from %s: %s', target, what, by, contents)
        super().on_ctcp(by, target, what, contents)

    async def on_ctcp_action(self, by, target, contents):
        await self.network_.on_ctcp_action(target, by, contents)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  Handling for rooms.
"""

import asyncio
import enum
import logging
import time as Time

import gi
gi.require_version('Notify', '0.7')
from gi.repository import Notify

from .widgets.message_buffer import MessageBuffer
from .widgets.message_row import MessageRow
from .widgets.room_row import RoomRow, RoomKind
from .widgets.topic import TopicPane

class Room:
    """ A Room object that represents a list of grouped messages on IRC. This 
    can be a channel, PM conversation/dialog/query, or a list of 
    server messages.
    
    Attributes:
        buffer (:obj:): The message buffer for this room.
        topic (:obj:): The room topic (if applicable).
        row (:obj:`RoomRow`) The row for this room in the switcher.
    """

    def __init__(self, app, network, window, name):
        Notify.init('AD1459')
        self.log = logging.getLogger('ad1459.room')
        self.log.debug('Creating new room %s', name)

        self.app = app
        self.network = network 
        self.window = window

        self.name = name

        self.notification = Notify.Notification.new('AD1459', 'Init')
        self.buffer = MessageBuffer(self)
        self.row = RoomRow(self)
        self.topic_pane = TopicPane(self)
        self.old_users = []
        self._tab_complete = []
    
    # Methods
    def add_message(self, message, sender='*', kind='message', time=None):
        """ Adds a message into this room and inserts it into the buffer.

        Arguments:
            message (str): the text of the message to add.
            sender (str): The person/entity who sent the message
            time (str): The time the message was sent.
            kind (str): The type of message this is (default: 'message')
        """
        if not time:
            time = Time.ctime().split()[3]
        
        if (
                self.network.nickname in message and 
                kind != 'action' and
                kind != 'server'
        ):
            kind = 'highlight'
        
        if sender == self.network.nickname and kind != 'server':
            kind = 'mine'

        new_message = MessageRow()
        new_message.kind = kind
        new_message.time = time
        new_message.sender = sender
        new_message.text = message

        ur_icon = self.row.unread_indicator.get_icon_name()[0]
        current_room = self.window.message_stack.get_visible_child().room
        
        if current_room != self or not self.window.focused:
            if kind != 'server':
                if self.network.nickname in message:
                    self.row.set_icon('emblem-important-symbolic')
                
                elif ur_icon != 'emblem-important-symbolic':
                    self.row.set_icon('radio-checked-symbolic')
            
            elif ur_icon != 'emblem-important-symbolic':
                self.row.set_icon('radio-mixed-symbolic')
        
        self.log.debug('Window is focused: %s', self.window.focused)
        # TODO: This can be improved
        if not self.window.focused:
            if (
                    new_message.kind == 'highlight' or 
                    (self.kind == RoomKind.DIALOG and
                    new_message.kind != 'mine')
            ):
                if self.kind == RoomKind.CHANNEL:
                    self.notification.update(
                        f'{sender} mentioned you in {self.name}',
                        message
                    )
                
                else:
                    self.notification.update(
                        f'New message from {sender}',
                        message
                    )
                self.notification.show()
        
        messages = self.buffer.list_box.get_children()
        if messages:
            last_message = messages[-1]
            if last_message:
                if kind == 'server' and last_message.kind == 'server':
                    last_message.text += f' {message}'
                    last_message.time = time
                
                else:
                    self.buffer.add_message_to_buffer(new_message)

            else:
                self.buffer.add_message_to_buffer(new_message)
        
        else:
            self.buffer.add_message_to_buffer(new_message)
        
        self.window.show_all()
    
    def update_tab_complete(self, user):
        if user in self._tab_complete:
            self._tab_complete.remove(user)
        
        self._tab_complete.append(user)
    
    def update_users(self):
        self.old_users = []
        for user in self.users:
            self.old_users.append(user)
            
        self.topic_pane.update_users()
    
    def leave(self):
        """ Remove this room from the UI."""
        self.log.debug('Removing room %s form list', self.name)
        self.buffer.destroy()
        self.topic_pane.destroy()
        self.row.destroy()
        self.notification.close()
    
    def part(self):
        """ Sends a part from this channel."""
        asyncio.run_coroutine_threadsafe(
            self.network.client.part(self.name, message='Leaving...'),
            loop=asyncio.get_event_loop()
        )


    # Data
    @property
    def data(self):
        """dict: A dictionary with this channel's data."""
        try:
            return self.network.client.channels[self.name]
        except KeyError:
            return {
                'users': {self.name, self.network.nickname},
                'topic': f'Private chat with {self.name}',
                'topic_by': self.network.name,
                'topic_set': Time.ctime().split()[3]
            }
    
    @property
    def users(self):
        """list: A list of users in this room."""
        users = []

        if self.kind == RoomKind.SERVER:
            users.append(self.network.nickname)
        
        elif self.kind == RoomKind.DIALOG:
            users.append(self.name)
        
        else:
            for user in self.data['users']:
                users.append(user)

        return users
    
    @property
    def tab_complete(self):
        return self._tab_complete

    @property
    def kind(self):
        """:obj:`RoomKind` The type of room this is."""
        return self._kind
    
    @kind.setter
    def kind(self, kind):
        self.log.debug('Setting room type for %s to %s', self.name, kind)
        kind = kind.upper()
        self._kind = RoomKind[kind]
        self.row.set_margins()
        self.log.debug('Set margins for %s room', kind)

    @property
    def name(self):
        """str: The name of this channel. This is displayed in the UI, and is
        the name of the channel, the network, or the user we are chatting with.
        """
        return self._name
    
    @name.setter
    def name(self, name):
        self._name = name
    
    @property
    def id(self):
        """str: The internal ID of this channel. This is used internally to 
        identify this room and it's children within the UI.

        We use the id() function to get a unique identifier for self, since this
        is guaranteed to be unique while the object exists.
        """
        return f'{self.name}-{id(self)}'



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This is the main module. When imported, it should run the application and 
  spawn the GUI.
"""

import asyncio
import logging
import pydle
import threading

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, Gio

# from .window import AdWindow
        

# app = Gtk.Application.new(
#   'in.donotspellitgav.ad1459', Gio.ApplicationFlags.FLAGS_NONE
# )

# print('Setting up logging')
# log = logging.getLogger('ad1459')
# handler = logging.StreamHandler()
# formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
# handler.setFormatter(formatter)
# log.addHandler(handler)
# log.setLevel(logging.DEBUG)
# log.debug('Initializing application')

# app.nick = "ad1459"
# app.username = 'ad1459'
# app.realname = 'AD1459 User'

# app.window = AdWindow(app)
# app.window.set_default_size(1000,600)
# app.window.connect('delete-event', Gtk.main_quit)
# app.window.show_all()

# log.debug('Initializing IRC thread')
# irc = threading.Thread(target=asyncio.get_event_loop().run_forever)
# irc.daemon = True
# irc.start()

# log.debug('Starting GTK Main Loop')
# Gtk.main()

# app.run()


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  ListBoxRows for networks/rooms.
"""

from enum import Enum

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class RoomKind(Enum):
    """ An enum to classify the type of room this is.
    """
    SERVER = 1
    CHANNEL = 2
    ROOM = 2
    PRIVMSG = 3
    QUERY = 3
    DIALOG = 3
    WHISPER = 3

    def __str__(self):
        """ Turn this back into a string. """
        strings = {
            1: 'network',
            2: 'channel',
            3: 'user'
        }
        return strings[self.value]

def sort_by_server(row1, row2):
    if row1.room.network.name < row2.room.network.name:
        return -1
    
    else:
        return 1

def sort_by_room(row1, row2):
    if row1.room.name < row2.room.name:
        return -1
    
    else:
        return 1

def room_row_sort(row1, row2, *user_data):
    """ Tells whether row1 should be before row2 or not

    Arguments:
        row1 (:obj:`RoomRow`): The first row to compare
        row2 (:obj:`RoomRow`): The second row to compare
        user_data (:obj:`object or None): user data
    
    Returns:
        -1 if row1 should go before row2.
        0  if they are equal and the order doesn't matter. 
        1  if row1 should go after row2.
    """
    
    if row1.kind == RoomKind.SERVER:
        if row2.kind == RoomKind.SERVER:
            return sort_by_server(row1, row2)
        else:
            
            if row2.room in row1.room.network.rooms:
                return -1

            else:
                return sort_by_server(row1, row2)

    elif row1.kind == RoomKind.CHANNEL:
        if row2.kind == RoomKind.SERVER:
            if row1.room in row2.room.network.rooms:
                return 1
            
            else:
                return sort_by_server(row1, row2)

        elif row2.kind == RoomKind.CHANNEL:
            if row1.room in row2.room.network.rooms:
                return sort_by_room(row1, row2)
            
            else:
                return sort_by_server(row1, row2)

        elif row2.kind == RoomKind.DIALOG:
            if row1.room in row2.room.network.rooms:
                return -1
            
            else:
                return sort_by_server(row1, row2)

    elif row1.kind == RoomKind.DIALOG:
        if row2.kind == RoomKind.SERVER or row2.kind == RoomKind.CHANNEL:
            if row1.room in row2.room.network.rooms:
                return 1

            else:
                
                return sort_by_server(row1, row2)

        if row2.kind == RoomKind.DIALOG:
            if row1.room in row2.room.network.rooms:
                return sort_by_room(row1, row2)
            
            else:
                
                return sort_by_server(row1, row2)

class RoomRow(Gtk.ListBoxRow):
    """ A dedicated ListBoxRow for representing a room in the Room List.

    Arguments:
        kind (:obj:`RoomKind`): The type of room this is.
    
    Attributes:
        room_name (str): The name of this room.
        kind (RoomKind Enum): The type of room this is.
    """

    def __init__(self, room):
        super().__init__()
        self.room = room
        self.set_can_focus(False)

        grid = Gtk.Grid()
        grid.set_column_spacing(6)
        grid.set_margin_top(3)
        grid.set_margin_bottom(3)
        grid.set_margin_start(3)
        grid.set_margin_end(3)
        self.add(grid)

        self.room_label = Gtk.Label()
        self.room_label.set_text(self.room.name)
        grid.attach(self.room_label, 0, 0, 1, 1)

        self.unread_indicator = Gtk.Image.new_from_icon_name(
            'radio-symbolic',
            Gtk.IconSize.SMALL_TOOLBAR            
        )
        grid.attach(self.unread_indicator, 1, 0, 1, 1)

    # Methods
    def set_margins(self):
        """Sets the margins on self depending on what type of room we are."""
        print(self.room.kind)
        if self.room.kind == RoomKind.SERVER:
            self.set_margin_top(12) 
            self.set_margin_start(0)
        elif self.room.kind == RoomKind.CHANNEL: 
            self.set_margin_top(0)
            self.set_margin_start(12)
        elif self.room.kind == RoomKind.DIALOG:
            self.set_margin_top(0)
            self.set_margin_start(18)
        else:
            self.set_margin_top(0)
            self.set_margin_start(0)

    def set_icon(self, icon_name):
        self.unread_indicator.set_from_icon_name(
            icon_name, Gtk.IconSize.SMALL_TOOLBAR
        )

    # Data
    @property
    def kind(self):
        """Easier compatibility with sort func."""
        return self.room.kind
   


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  A parser for IRC style tags.
"""

import logging
from urllib.parse import urlparse

class Parser:

    formatting = {
        '\x02': ['b', '*'],
        #'\u0003': 'color',
        #'\u000F': 'clear',
        '\x1D': ['i', '_'],
        '\x1F': ['u', '-']
    }

    def __init__(self):
        self.log = logging.getLogger('ad1459.formatting')
    
    def format_text(self, text):
        for i in self.formatting:
            # text = text.replace(f'/{self.formatting[i][0]}', i)
            text = text.replace(f'/{self.formatting[i][1]}', i)
        return text
    
    def sanitize_message(self, message):
        """ Strips IRC formatting from a message to make it suitable for display
        outside of the client, e.g. in notifications.

        Arguments:
            message (str): The message to clean up.

        Returns:
            str: The cleaned message.
        """
        formatting_chars = [
            '\x02', # bold
            '\x1D', # italics
            '\x1F', # underline
            '\x1E', # strikethrough
            '\x11'
        ]

        for fchar in formatting_chars:
            message = message.replace(fchar, '')
        
        return message

    def parse_text(self, text):        
        # \u0002 bold
        # \u0003 colour
        # \u000F cancel all
        # \u001D italic
        # \u001F underline
        text = self.fix_markedup_tags(text)
        f_text = ''
        current_tags = []

        for char in text:
            if char in self.formatting:

                if not char in current_tags:
                    for tag in reversed(current_tags):
                        f_text = f'{f_text}</{self.formatting[tag][0]}>'

                    current_tags.append(char)
                    for tag in current_tags:
                        f_text = f'{f_text}<{self.formatting[tag][0]}>'

                else:
                    for tag in reversed(current_tags):
                        f_text = f'{f_text}</{self.formatting[tag][0]}>'

                    current_tags.pop()
                    for tag in current_tags:
                        f_text = f'{f_text}<{self.formatting[tag][0]}>'

            else: 
                f_text = f'{f_text}{char}'
        
        for tag in reversed(current_tags):
            f_text = f'{f_text}</{self.formatting[tag][0]}>'

        return f_text
    
    def hyperlinks(self, text):
        words = text.split()

        linked_words = []
        for word in words:
            scheme = urlparse(word).scheme
            if scheme == 'http' or scheme == 'https':
                paren = False
                if word.endswith(')'):
                    word = word.strip(')')
                    paren = True
                word = f'<a href="{word}">{word}</a>'
                if paren:
                    word = f'{word})'
            linked_words.append(word)
        return " ".join(linked_words)

    def fix_markedup_tags(self, text):
        mu_formatting = {
            '&#x2;': '\x02',
            #'\u0003': #'\u0003',
            #'\u000F': #'\u000F',
            '&#x1d;': '\x1D',
            '&#x1f;': '\x1F'
        }

        for i in mu_formatting:
            text = text.replace(i, mu_formatting[i])
        
        return text


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  Main Application
"""

import asyncio
import logging
import pathlib
import os
import threading

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Gdk': '3.0'
    }
)
from gi.repository import Gtk, Gdk, Gio

from .formatting import Parser
from .network import Network
from .widgets.window import Ad1459Window
from . import handlers

USER_HOME_PATH = str(pathlib.Path.home())
CONFIG_DIR_PATH = os.path.join(
    USER_HOME_PATH, 
    os.path.join('.config', 'ad1459')
)
CONFIG_FILE_PATH = os.path.join(CONFIG_DIR_PATH, 'servers.ini')

class Ad1459Application:

    def __init__(self):
        self.log = logging.getLogger('ad1459')
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self.log.addHandler(handler)
        self.log.setLevel(logging.DEBUG)
        self.log.debug('Initializing application')

        self.user_home_path = USER_HOME_PATH
        self.config_dir_path = CONFIG_DIR_PATH
        self.config_file_path = CONFIG_FILE_PATH

        self.windows = []
        self.networks = []

        self.app = Gtk.Application.new(
            'in.donotspellitgav.in', Gio.ApplicationFlags.FLAGS_NONE
        )
        self.app.connect('activate', self.init_application)
        self.parser = Parser()

                # Set up CSS
        css = (
            b'.message-row {'
            b'  border-radius: 6px;'
            b'  background-color: alpha(@theme_selected_bg_color, 0.1);'
            b'}'
            b'.mine {'
            b'  background-color: alpha(@theme_selected_bg_color, 0.2);'
            b'}'
            b'.highlight {'
            b'    background-color: alpha(@success_color, 0.3);'
            b'}'
            b'.server {'
            b'    background-color: transparent;'
            b'}'
            b'.notice {'
            b'    background-color: alpha(@warning_color, 0.3);'
            b'}'
            b'.connect-entry {'
            b'    background-color: @theme_base_color;'
            b'}'
            b'.topic-expander {'
            b'    font-size: 0.75em;'
            b'}'
        )

        screen = Gdk.Screen.get_default()
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(css)
        style_context = Gtk.StyleContext()
        style_context.add_provider_for_screen(
            screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def init_application(self):
        """ Starts up all of the application loops and peices."""
        self.log.debug('Initializing IRC thread')
        self.irc = threading.Thread(target=asyncio.get_event_loop().run_forever)
        self.irc.daemon = True
        self.irc.start()

        self.log.debug('Creating initial window...')
        first_window = self.init_window()
        self.windows.append(first_window)
        first_window.show_all()

        self.log.debug('Starting up GTK main loop.')
        Gtk.main()

    def init_window(self):
        """ Create a a window for the application.
        
        Returns:
            A Gtk.Window set up for AD1459
        """
        self.log.debug('Adding window')
        window = Ad1459Window(self, self.parser)
        window.set_default_size(1000, 600)
        window.connect('delete-event', self.remove_window)
        self.connect_ui(window)
        
        window.show_all()
        return window

    def connect_ui(self, window):
        """ Connects the UI in window to the correct signal handlers.

        Arguments:
            window (:obj:`Ad1459Windw`): The window object to connect.
        """
        # Hook up the IRC Entry and related buttons
        window.irc_entry.connect(
            'activate', 
            handlers.on_send_button_clicked,
            window.irc_entry.get_text(),
            window.switcher.get_active_room(),
            window
        )
        window.send_button.connect(
            'clicked', 
            handlers.on_send_button_clicked,
            window.switcher.get_active_room(),
            window
        )
        window.nick_button.connect(
            'clicked',
            handlers.on_nick_button_clicked,
            window
        )

        # Hook up header functions
        window.header.close_button.connect(
            'clicked',
            handlers.on_appmenu_close_clicked,
            window.switcher.get_active_room(),
            window
        )
        window.header.about_button.connect(
            'clicked',
            handlers.on_appmenu_about_clicked,
            window
        )

        # ServerPopup functions
        window.header.server_popup.connect_button.connect(
            'clicked',
            handlers.on_server_popup_connect_clicked,
            window
        )

        # Channel Switcher 
        window.switcher.switcher.connect(
            'row-selected',
            handlers.on_room_selected,
            window
        )
        window.join_entry.connect(
            'activate',
            handlers.on_join_entry_activate,
            window
        )

    def remove_window(self, window, data=None):
        """ Deletes a window and moves all of its stuff to another window.

        Arguments: 
            window (Gtk.Window): The window to remove.
        """
        self.log.debug('Deleting window')
        window.destroy()

        if len(self.windows) <= 1:
            self.log.debug('Last window destroyed, quitting.')
            Gtk.main_quit()
            exit(0)

    # def add_network(
    #         self,
    #         name=None,
    #         auth=None,
    #         host=None,
    #         port=None,
    #         tls=True,
    #         nick=None,
    #         user=None,
    #         real=None,
    #         pasw=None
    # ):
    #     """ Adds a network object to this application.

    #     Returns:
    #         A :obj:`Network` for the new network.
    #     """
    #     new_network = Network(self)
        
    #     if name:
    #         new_network.name = name

    #     if auth:
    #         new_network.auth = auth

    #     if host:
    #         new_network.host = host

    #     if port:
    #         new_network.port = port

    #     if not tls:
    #         new_network.tls = tls

    #     if nick:
    #         new_network.nickname = nick

    #     if user:
    #         new_network.username = user

    #     if real:
    #         new_network.realname = real

    #     if pasw:
    #         new_network.password = pasw



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  Signal handlers for UI components.
"""

import logging

from .network import Network
from .widgets.room_row import RoomKind
from .widgets.about import AboutDialog

def on_send_button_clicked(widget, text, room, window, data=None):
    """ `clicked` signal handler for the Send button.

    Also handles the `activate` signal for the IRC Entry.
    
    Arguments:
        text (str): The text to send to IRC.
        room (:obj:`Room`): The room to send `text` to.
        window (:obj:`Ad1459Application`): The window we're in
    """
    room = window.message_stack.get_visible_child().room
    network = room.network
    message = window.irc_entry.get_text()
    if message:
        network.send_message(room, message)
    window.irc_entry.set_text('')

def on_nick_button_clicked(button, window, data=None):
    """`clicked` signal handler for the nickname button.
    
    Arguments:
        text (str): The text to set the nickname to.
        network (:obj:`Network`): The network whose nick to set.
        window (:obj:`Ad1459Application`): The window we're in
    """
    network = window.active_room.network
    new_nick = window.irc_entry.get_text()
    network.change_nick(new_nick)
    window.irc_entry.set_text('')

def on_server_popup_connect_clicked(button, window, data=None):
    """`clicked` signal handler for the Connect button in the server popup
    
    Arguments:
        window (:obj:`Ad1459Application`): The window we're in
    """
    popup = window.header.server_popup
    app = window.app
    network_line = popup.server_line
    window.header.spinner.start()
    
    if network_line:
        pass

    else:
        network = Network(app, window)
        network.name = popup.name
        network.auth = popup.auth
        network.host = popup.server
        network.port = popup.port
        network.tls = popup.tls
        network.nickname = popup.nick
        network.username = popup.username
        network.realname = popup.realname
        network.password = popup.password
    
    app.networks.append(network)
    
    if popup.save:
        popup.save_details()

    network.connect()
    popup.popdown()
    popup.layout_grid.set_sensitive(False)

def on_appmenu_close_clicked(button, room, window, data=None):
    """`clicked` signal handler for the Leave button in the appmenu.
    
    Arguments:
        room (:obj:`Room`): The Room to leave.
        window (:obj:`Ad1459Application`): The window we're in
    """
    room = window.message_stack.get_visible_child().room
    network = room.network
    if room.kind == RoomKind.CHANNEL:
        room.part()
    elif room.kind == RoomKind.SERVER:
        for room in network.rooms:
            room.leave()
        room.network.disconnect()
        window.app.networks.remove(network)
        window.switcher.invalidate_sort()
    else:
        room.network.rooms.remove(room)
        room.leave()
        window.switcher.invalidate_sort()


def on_appmenu_about_clicked(button, window, data=None):
    """`clicked` signal handler for the About button.
    
    Arguments:
        window (:obj:`Ad1459Application`): The window we're in
    """
    about_dialog = AboutDialog()
    about_dialog.run()
    about_dialog.destroy()

def on_room_selected(listbox, row, window, data=None):
    """`row-selected` signal handler for room switcher.

    Arguments:
        window (:obj:`Gtk.Window`): The window we're part of.
    """
    try:
        row.room.update_users()
        row.room.topic_pane.update_topic()
        row.room.notification.close()
        window.show_all()
        row.set_icon('radio-symbolic')
        window.message_stack.set_visible_child_name(row.room.id)
        window.topic_stack.set_visible_child_name(row.room.id)
        window.irc_entry.grab_focus_without_selecting()
        window.nick_button.set_label(row.room.network.nickname)
        
        if row.room.kind == RoomKind.DIALOG:
            window.header.close_button.set_label('Close conversation')
        
        elif row.room.kind == RoomKind.SERVER:
            window.header.close_button.set_label('Disconnect from server')
        
        elif row.room.kind == RoomKind.CHANNEL:
            window.header.close_button.set_label('Leave channel')
    
    except AttributeError:
        return
    
    window.switcher.invalidate_sort()

def on_join_entry_activate(entry, window, data=None):
    """`activate` signal handler for join entry.

    also handles the `icon-release` signal.
    """
    room_name = entry.get_text()
    network = window.message_stack.get_visible_child().room.network
    
    if room_name.startswith('#'):
        network.join_channel(room_name)
    
    else:
        room = network.get_room_for_name(room_name)

    
    entry.set_text('')
    




#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This file is an application window.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Gdk': '3.0'
    }
)
from gi.repository import Gtk, Gdk
from .room_row import room_row_sort

class RoomSwitcher(Gtk.Grid):

    def __init__(self, app):
        super().__init__()
        self.app = app

        window = Gtk.ScrolledWindow()
        window.set_hexpand(True)
        window.set_vexpand(True)
        self.attach(window, 0, 0, 1, 1)

        self.switcher = Gtk.ListBox()
        self.switcher.set_hexpand(True)
        self.switcher.set_vexpand(True)
        self.switcher.set_sort_func(room_row_sort)
        window.add(self.switcher)
    
    def add_row(self, row):
        """ Adds a row to the ListBox.

        Arguments:
            row (:obj:`RoomRow`): A Room Row to add to the listbox.
        """
        self.switcher.insert(row, 0)
    
    def remove_row(self, row):
        """ Removes a row from the ListBox.

        Arguments:
            row (:obj:`RoomRow`): The row to remove
        """
        try:
            row.destroy()

        except AttributeError:
            pass

        self.switcher.invalidate_sort()
    
    def invalidate_sort(self):
        """ proxy for self.switcher.invalidate_sort()"""
        self.switcher.invalidate_sort()
    
    def get_active_room(self):
        """Gets the currently selected Room.

        Returns:
            a :obj:`Room` for the currently selected row
        """
        room = self
        room.network = ""
        return room




#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  A popover for connecting to servers.
"""

import configparser
import keyring as Keyring
import os
import pathlib

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

class ServerPopover(Gtk.Popover):

    def __init__(self, config_file_path):
        super().__init__()

        self.config = configparser.ConfigParser()
        self.keyring = Keyring.get_keyring()

        self.config_file_path = config_file_path
        pathlib.Path(os.path.dirname(self.config_file_path)).mkdir(
            parents=True, exist_ok=True
        )

        self.config.read(self.config_file_path)
        
        self.layout_grid = Gtk.Grid()
        self.layout_grid.set_column_spacing(6)
        self.layout_grid.set_row_spacing(12)
        self.layout_grid.props.margin = 6
        self.add(self.layout_grid)
        
        # Saved Networks
        self.saved_combo = Gtk.ComboBoxText()
        self.init_saved_combo()
        self.saved_combo.set_active(-1)
        Gtk.StyleContext.add_class(
            self.saved_combo.get_style_context(), 'connect-entry'
        )
        self.saved_combo.connect('changed', self.on_saved_combo_changed)
        self.layout_grid.attach(self.saved_combo, 0, 0, 2, 1)

        self.connect_grid = Gtk.Grid()
        self.connect_grid.set_orientation(Gtk.Orientation.VERTICAL)
        Gtk.StyleContext.add_class(self.connect_grid.get_style_context(), 'linked')
        Gtk.StyleContext.add_class(self.connect_grid.get_style_context(), 'connect-entry')
        self.layout_grid.attach(self.connect_grid, 0, 1, 2, 1)

        # Network Name
        self.name_entry = Gtk.Entry()
        self.name_entry.simple_entry = True
        self.name_entry.set_placeholder_text('Network Name')
        self.connect_grid.attach(self.name_entry, 0, 0, 2, 1)

        # nickname
        self.nick_entry = Gtk.Entry()
        self.nick_entry.simple_entry = True
        self.nick_entry.set_placeholder_text('Nickname')
        self.connect_grid.attach(self.nick_entry, 0, 1, 2, 1)

        # username
        self.username_entry = Gtk.Entry()
        self.username_entry.simple_entry = True
        self.username_entry.set_placeholder_text('Username/Ident')
        self.connect_grid.attach(self.username_entry, 0, 2, 2, 1)

        # realname
        self.realname_entry = Gtk.Entry()
        self.realname_entry.simple_entry = True
        self.realname_entry.set_placeholder_text('Real Name')
        self.connect_grid.attach(self.realname_entry, 0, 3, 2, 1)

        # server
        self.server_entry = Gtk.Entry()
        self.server_entry.simple_entry = True
        self.server_entry.set_placeholder_text('Server')
        self.connect_grid.attach(self.server_entry, 0, 4, 1, 1)

        # port
        port_adj = Gtk.Adjustment.new(
            value=6697,
            lower=1,
            upper=65535,
            step_increment=1,
            page_increment=1,
            page_size=0
        )

        self.port_entry = Gtk.SpinButton()
        self.port_entry.simple_entry = True
        self.port_entry.configure(port_adj, 1, 0)
        self.port_entry.set_numeric(True)
        self.port_entry.set_placeholder_text('Port')
        self.connect_grid.attach(self.port_entry, 1, 4, 1, 1)

        # auth
        self.auth_combo = Gtk.ComboBoxText()
        self.auth_combo.simple_entry = True
        self.auth_combo.append_text('SASL Authentication')
        self.auth_combo.append_text('Server Password')
        self.auth_combo.append_text('No Authentication')
        self.auth_combo.set_active(0)
        Gtk.StyleContext.add_class(
            self.auth_combo.get_style_context(), 'connect-entry'
        )
        self.connect_grid.attach(self.auth_combo, 0, 5, 2, 1)

        # password
        self.password_entry = Gtk.Entry()
        self.password_entry.simple_entry = True
        self.password_entry.set_placeholder_text('Password')
        self.password_entry.set_visibility(False)
        self.connect_grid.attach(self.password_entry, 0, 6, 2, 1)

        # TLS Check
        self.tls_check = Gtk.CheckButton()
        self.tls_check.simple_entry = True
        self.tls_check.set_label('TLS')
        self.tls_check.set_active(True)
        self.layout_grid.attach(self.tls_check, 0, 2, 1, 1)

        # Save check
        self.save_check = Gtk.CheckButton()
        self.save_check.set_label('Save')
        self.save_check.set_active(True)
        self.layout_grid.attach(self.save_check, 1, 2, 1, 1)

        # Server entry
        self.server_line_entry = Gtk.Entry()
        self.server_line_entry.simple_entry = False
        self.server_line_entry.set_placeholder_text('Server line')
        self.layout_grid.attach(self.server_line_entry, 0, 3, 2, 1)

        # Connect Button
        self.connect_button = Gtk.Button()
        self.connect_button.set_label('Connect')
        Gtk.StyleContext.add_class(
            self.connect_button.get_style_context(), 'suggested-action'
        )
        self.layout_grid.attach(self.connect_button, 0, 4, 2, 1)

        self.widgets = [
            self.name_entry,
            self.nick_entry,
            self.username_entry,
            self.realname_entry,
            self.server_entry,
            self.password_entry
        ]

    @property 
    def name(self):
        return self.name_entry.get_text()

    @property 
    def nick(self):
        return self.nick_entry.get_text()

    @property 
    def username(self):
        if self.username_entry.get_text() != '':
            return self.username_entry.get_text()
        else:
            return self.nick

    @property 
    def realname(self):
        if self.realname_entry.get_text() != '':
            return self.realname_entry.get_text()
        else:
            return self.nick

    @property 
    def server(self):
        return self.server_entry.get_text()

    @property 
    def port(self):
        return self.port_entry.get_value()

    @property 
    def auth(self):
        auth = self.auth_combo.get_active_text()
        if auth == 'SASL Authentication':
            return 'sasl'

        elif auth == 'Server Password':
            return 'pass'

        else:
            return 'none'

    @property 
    def password(self):
        return self.password_entry.get_text()

    @property 
    def tls(self):
        return self.tls_check.get_active()

    @property 
    def save(self):
        return self.save_check.get_active()

    @property 
    def server_line(self):
        return self.server_line_entry.get_text()
    
    def on_saved_combo_changed(self, combo, data=None):
        """ changed signal handler for saved_combo."""
        network = combo.get_active_text()

        if network == 'New...':
            self.reset_all_text()
        
        else:
            self.config.read(self.config_file_path)
            self.name_entry.set_text(
                self.config[network]['name']
            )
            self.nick_entry.set_text(
                self.config[network]['nickname']
            )
            self.username_entry.set_text(
                self.config[network]['username']
            )
            self.realname_entry.set_text(
                self.config[network]['realname']
            )
            self.server_entry.set_text(
                self.config[network]['host']
            )
            self.port_entry.set_value(
                int(self.config[network]['port'])
            )
            if self.config[network]['auth'] == 'sasl':
                self.auth_combo.set_active(0)
            elif self.config[network]['auth'] == 'pass':
                self.auth_combo.set_active(1)
            else:
                self.auth_combo.set_active(2)
            if self.config[network]['tls'] == 'True':
                self.tls_check.set_active(True)
            else:
                self.tls_check.set_active(False)
            keyring_service = (
                f'{self.config[network]["name"]}-{self.config[network]["host"]}'
            )
            keyring_username = (
                f'{self.config[network]["nickname"]}'
                f'!{self.config[network]["username"]}'
                f'@{self.config[network]["name"]}'
            )
            self.password_entry.set_text(
                self.keyring.get_password(keyring_service, keyring_username)
            )
            self.save_check.set_active(True)

    def init_saved_combo(self):
        """Clears any existing items from the saved network combo and sets it 
        up with new ones.
        """
        self.saved_combo.remove_all()
        self.saved_combo.append_text('New...')
        for network in self.config.sections():
            self.saved_combo.append_text(network)
    
    def save_details(self):
        self.config[self.name] = {}
        self.config[self.name]['name'] = self.name
        self.config[self.name]['auth'] = self.auth
        self.config[self.name]['host'] = self.server
        self.config[self.name]['port'] = str(int(self.port))
        self.config[self.name]['tls'] = str(self.tls)
        self.config[self.name]['nickname'] = self.nick
        self.config[self.name]['username'] = self.username
        self.config[self.name]['realname'] = self.realname

        with open(self.config_file_path, mode='w') as configfile:
            self.config.write(configfile)
        
        keyring_username = f'{self.nick}!{self.username}@{self.name}'
        self.keyring.set_password(
            f'{self.name}-{self.server}',
            keyring_username,
            self.password
        )

    def reset_all_text(self):
        self.saved_combo.set_active(0)
        self.name_entry.set_text('')
        self.nick_entry.set_text('')
        self.username_entry.set_text('')
        self.realname_entry.set_text('')
        self.server_entry.set_text('')
        self.port_entry.set_value(6697)
        self.auth_combo.set_active(0)
        self.password_entry.set_text('')
        self.tls_check.set_active(True)
        self.save_check.set_active(True)
        self.server_line_entry.set_text('')

    def get_all_widgets_text(self):
        """ Returns the text of all widgets, formatted as a serverline."""

        line = ''
        if self.auth_combo.get_active_text() == 'SASL Authentication':
            line = 'sasl'
            
        elif self.auth_combo.get_active_text() == 'Server Password':
            line = 'pass'

        else:
            line = 'none'

        line = f'{line} {self.name_entry.get_text()}'
        line = f'{line} {self.server_entry.get_text()}'
        line = f'{line} {int(self.port_entry.get_value())}'
        line = f'{line} {self.nick_entry.get_text()}'
        line = f'{line} {self.username_entry.get_text()}'
        line = f'{line} {self.realname_entry.get_text()}'
        line = f'{line} {self.password_entry.get_text()}'
        
        return line.strip()



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This is the headerbar.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from .server_popup import ServerPopover

class Headerbar(Gtk.HeaderBar):
    """ The headerbar we use."""

    def __init__(self, app):
        super().__init__()

        self.set_show_close_button(True)
        self.set_title('AD1459')
        self.set_has_subtitle(False)

        self.network_button = Gtk.MenuButton()
        network_image = Gtk.Image.new_from_icon_name(
            'network-server-symbolic',
            Gtk.IconSize.BUTTON
        )
        self.network_button.set_image(network_image)
        self.pack_start(self.network_button)

        self.server_popup = ServerPopover(app.config_file_path)
        self.server_popup.show_all()
        self.server_popup.popdown()
        self.network_button.set_popover(self.server_popup)

        self.spinner = Gtk.Spinner()
        self.pack_start(self.spinner)

        self.appmenu = Gtk.Popover()
        appmenu_grid = Gtk.Grid()
        appmenu_grid.set_margin_start(6)
        appmenu_grid.set_margin_end(6)
        appmenu_grid.set_margin_top(6)
        appmenu_grid.set_margin_bottom(6)
        self.appmenu.add(appmenu_grid)

        self.close_button = Gtk.ModelButton()
        self.close_button.set_label('Close conversation')
        appmenu_grid.attach(self.close_button, 0, 0, 1, 1)

        self.about_button = Gtk.ModelButton()
        self.about_button.set_label('About AD1459')
        appmenu_grid.attach(self.about_button, 0, 1, 1, 1)

        self.appmenu.show_all()
        self.appmenu.popdown()

        appmenu_button = Gtk.MenuButton()
        appmenu_image = Gtk.Image.new_from_icon_name(
          'view-more-symbolic',
          Gtk.IconSize.BUTTON
        )
        appmenu_button.set_image(appmenu_image)
        appmenu_button.set_popover(self.appmenu)
        self.pack_end(appmenu_button)



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  A buffer full of messages.
"""

import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class MessageBuffer(Gtk.ScrolledWindow):

    def __init__(self, room):
        super().__init__()
        self.room = room

        self.list_box = Gtk.ListBox()
        self.list_box.set_vexpand(True)
        self.list_box.set_hexpand(True)
        self.list_box.connect('size-allocate', self.scroll_to_bottom)

        self.adj = self.get_vadjustment()
        self.adj.connect('value-changed', self.on_window_scroll)
        self.add(self.list_box)

    def add_message_to_buffer(self, message):
        """Adds a message into the buffer."""
        self.list_box.add(message)

        # Scroll to the bottom if this is our own message.
        if message.sender == self.room.network.nickname:
            self.scroll_to_bottom(self.list_box)
    
    def on_window_scroll(self, widget, data=None):
        """value-changed signal handler for the window adjustment."""
        adj = self.get_vadjustment()
        max_value = adj.get_upper() - adj.get_page_size()

        if adj.get_value() < max_value:
            try:
                self.list_box.disconnect_by_func(self.scroll_to_bottom)
            
            except TypeError:
                pass
        
        else:
            try:
                self.list_box.disconnect_by_func(self.scroll_to_bottom)
            
            except TypeError:
                pass

            self.list_box.connect('size-allocate', self.scroll_to_bottom)

    def scroll_to_bottom(self, widget, data=None):
        """Scroll to the bottom of the buffer."""
        adj = self.get_vadjustment()
        max_value = adj.get_upper() - adj.get_page_size()
        adj.set_value(max_value)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  ListBoxRows for networks/rooms.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class UserRow(Gtk.ListBoxRow):

    def __init__(self, room):
        super().__init__()
        self.room = room
        grid = Gtk.Grid()
        grid.set_column_spacing(6)
        grid.set_margin_top(3)
        grid.set_margin_bottom(3)
        grid.set_margin_start(3)
        grid.set_margin_end(3)
        self.add(grid)

        self.status_image = Gtk.Image.new_from_icon_name(
            'radio-symbolic',
            Gtk.IconSize.SMALL_TOOLBAR
        )
        self.status_image.props.opacity = 0
        grid.attach(self.status_image, 0, 0, 1, 1)

        self.user_label = Gtk.Label()
        grid.attach(self.user_label, 1, 0, 1, 1)
    
    # Data
    @property
    def nick(self):
        """str: The user's nickname."""
        return self.user_label.get_text()
    
    @nick.setter
    def nick(self, nick):
        self.user_label.set_text(nick)

        


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  Handling for networks and their rooms.
"""

import asyncio
import logging
import time

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib

from .widgets.room_row import RoomRow
from .widgets.message_row import MessageRow
from .client import Client
from .commands import Commands
from .formatting import Parser
from .room import Room

class Network:
    """An IRC network to connect to and chat on.

    Attributes:
        rooms (list of :obj:`Room`): A list of the joined rooms on this network.
        name (str): The user-defined name for this network.
        auth (str): The user-authentication for this network. 'sasl', 'pass', 
            or 'none'.
        host (str): The hostname for the server to connect to.
        port (int): The port number to use.
        tls (bool): Wheter the connection uses TLS encryption.
        nickname (str): The user's nickname on this network.
        realname (str): The user's Real Name on this network.
        password (str): The authentication for the password on this network.
        app (:obj:`Ad1459Application`): The app we're running on.
    """
    cmd = Commands()
    commands = {
        '/me': cmd.me
    }

    def __init__(self, app, window):
        self.log = logging.getLogger('ad1459.network')
        self.log.debug('Creating network')
        self.app = app
        self.parser = Parser()
        self.window = window
        self.rooms = []
        self._config = {
            'name': 'New Network',
            'auth': 'sasl',
            'host': 'chat.freenode.net',
            'port': 6697,
            'tls': True,
            'nickname': 'ad1459-user',
            'username': 'ad1459-user',
            'realname': 'AD1459 User',
            'password': 'hunter2'
        }
        self.client = None

    # Synchronous Methods for this object.
    def connect(self):
        """ Connect to the network, disconnecting first if already connected. """
        if self.auth == 'sasl':
            self.client = Client(
                self.nickname, 
                self, 
                sasl_password=self.password, 
                sasl_username=self.username
            )
        else:
            self.client = Client(self.nickname, self)
        
        self.client.username = self.username

        self.log.debug('Spinning up async connection to %s', self.host)

        self.server_room = Room(self.app, self, self.window, self.name)
        self.server_room.kind = "server"
        self.add_room(self.server_room)

        if self.auth == 'pass':
            self.log.debug('Using password authentication')
            asyncio.run_coroutine_threadsafe(
                self.client.connect(
                    hostname=self.host,
                    port=self.port,
                    tls=self.tls,
                    password=self.password
                ),
                loop=asyncio.get_event_loop()
            )

        else:
            self.log.debug('Using SASL authentication (or none)')
            asyncio.run_coroutine_threadsafe(
                self.client.connect(
                    hostname=self.host,
                    port=self.port,
                    tls=self.tls
                ),
                loop=asyncio.get_event_loop()
            )
    
    def disconnect(self):
        """Disconnect from this network."""
        self.log.debug('Disconnecting from %s', self.name)
        asyncio.run_coroutine_threadsafe(
            self.client.quit(message='AD1459 Quit'),
            loop=asyncio.get_event_loop()
        )
    
    def change_nick(self, new_nick):
        """ Changes the user's nick."""
        asyncio.run_coroutine_threadsafe(
            self.client.set_nickname(new_nick),
            loop=asyncio.get_event_loop()
        )

    def join_channel(self, channel):
        """ Joins a room on the network."""
        asyncio.run_coroutine_threadsafe(
            self.client.join(channel), loop=asyncio.get_event_loop()
        )

    def add_room(self, room):
        """Adds a room to the window for this network."""
        self.log.debug('Adding room %s to the window', room.name)
        self.window.switcher.add_row(room.row)
        self.window.message_stack.add_named(room.buffer, room.id)
        self.window.topic_stack.add_named(room.topic_pane, room.id)
        self.window.show_all()
        self.window.switcher.invalidate_sort()
        self.rooms.append(room)
    
    def get_room_for_name(self, name):
        """ Gets a room object given the name."""
        for room in self.rooms:
            if room.name == name:
                return room

        if not name.startswith('#'):
            new_room = Room(self.app, self, self.window, name)
            new_room.kind = 'dialog'
            new_room.name = name
            new_room.topic_pane.update_topic()
            self.add_room(new_room)
            self.window.switcher.invalidate_sort()
            return new_room
    
    def send_message(self, room, message):
        """ Sends a message to the entity for room."""
        message_text = self.parser.format_text(message)
        for command in self.commands:
            if not message.startswith(command):
                asyncio.run_coroutine_threadsafe(
                    self.client.message(room.name, message_text),
                    loop=asyncio.get_event_loop()
                )
            else:
                self.commands[command](room, self.client, message_text)

    
    # Asynchronous Callbacks
    async def on_connected(self):
        """ Called upon connection to IRC."""
        self.log.info('Connected to %s', self.name)
        GLib.idle_add(self.do_connected)
    
    def do_connected(self):
        popup = self.window.header.server_popup
        popup.reset_all_text()
        popup.layout_grid.set_sensitive(True)
        self.window.header.spinner.stop()
        self.window.switcher.invalidate_sort()
    
    async def on_nick_change(self, old, new):
        self.log.debug('Nick %s changed to %s', old, new)
        GLib.idle_add(self.do_nick_change, old, new)
    
    def do_nick_change(self, old, new):
        if old == self.nickname or old == '<unregistered>':
            self.nickname = new
            self.window.nick_button.set_label(self.nickname)
        
        else:
            for room in self.rooms:
                if new in room.users:
                    room.add_message(f'{old} is now {new}.', kind='server')
                    room.window.show_all()
                    room.update_users()
    
    async def on_join(self, channel, user):
        self.log.debug('%s has joined %s', user, channel)
        GLib.idle_add(self.do_join, channel, user)
    
    def do_join(self, channel, user):
        if user == self.nickname:
            new_channel = Room(self.app, self, self.window, channel)
            new_channel.kind = 'channel'
            new_channel.name = channel
            new_channel.topic_pane.update_topic()
            self.add_room(new_channel)
            new_channel.update_users()
            self.window.switcher.invalidate_sort()
        
        else:
            room = self.get_room_for_name(channel)
            room.add_message(f'{user} has joined.', kind='server')
            self.window.show_all()
            room.update_users()
    
    async def on_part(self, channel, user, message=None):
        self.log.debug('%s has left %s, (%s)', user, channel, message)
        GLib.idle_add(self.do_part, channel, user, message)
    
    def do_part(self, channel, user, message=None):
        room = self.get_room_for_name(channel)
        if user == self.nickname:
            room.leave()
            room.network.server_room.add_message(
                f'You have left {room.name}', kind='server'
            )
            self.rooms.remove(room)
            self.window.switcher.invalidate_sort()
        
        else:
            room.update_users()
            room.add_message(f'{user} has left ({message}).', kind='server')
            self.window.show_all()
    
    async def on_kick(self, channel, target, by, reason=None):
        GLib.idle_add(self.do_kick, channel, target, by, reason)
    
    def do_kick(self, channel, target, by, reason=None):
        room = self.get_room_for_name(channel)
        if target == self.nickname:
            room.leave()
            room.network.server_room.add_message(
                f'You were kicked from {room.name} by {by}, reason: "{reason}"', 
                kind='server'
            )
            room.network.server_room.row.set_icon('emblem-important-symbolic')
            self.rooms.remove(room)
            self.window.switcher.invalidate_sort()
        
        else:
            room.update_users()
            room.add_message(
                f'{target} was kicked by {by}, reason: "{reason}".', kind='server'
            )
            self.window.show_all()

    async def on_quit(self, user, message=None):
        self.log.debug('%s has quit! (%s)', user, message)
        GLib.idle_add(self.do_quit, user, message)
    
    def do_quit(self, user, message=None):
        qmessage = f'{user} has quit ({message}).'
        for room in self.rooms:
            if user in room.old_users:
                room.add_message(qmessage, kind='server')
                room.update_users()
    
    async def on_kill(self, target, by, reason=None):
        GLib.idle_add(self.do_kill, target, by, reason)
    
    def do_kill(self, target, by, reason=None):
        qmessage = f'{target} was killed by {by}, reason: "{reason}".'
        for room in self.rooms:
            if target in room.old_users:
                room.add_message(qmessage, kind='server')
                room.update_users()
    
    async def on_message(self, target, source, message):
        GLib.idle_add(self.do_message, target, source, message)
    
    def do_message(self, target, source, message):
        if target.startswith('#'):
            room = self.get_room_for_name(target)
            self.log.debug('Adding message to %s', room.id)
            room.add_message(message, source)
            room.update_tab_complete(source)
            self.window.show_all()

    async def on_notice(self, target, source, message):
        self.log.debug('%s noticed to %s: %s', source, target, message)
        GLib.idle_add(self.do_notice, target, source, message)
    
    def do_notice(self, target, source, message):
        if target.startswith('#'):
            room = self.get_room_for_name(target)
            self.log.debug('Adding notice to %s', room.id)
            room.add_message(message, source, kind='notice')
            room.update_tab_complete(source)
            self.window.show_all()
    
    async def on_private_message(self, target, source, message):
        self.log.debug('PM to %s from %s: %s', target, source, message)
        GLib.idle_add(self.do_private_message, target, source, message)

    def do_private_message(self, target, source, message):
        if target == self.nickname:
            room = self.get_room_for_name(source)
            self.log.debug('Adding message from %s', room.id)
        else:
            room = self.get_room_for_name(target)
            self.log.debug('Adding message to %s', room.id)
        
        room.add_message(message, source)
        self.window.show_all()
    
    async def on_private_notice(self, target, source, message):
        self.log.debug('Private Notice to %s from %s: %s', target, source, message)
        GLib.idle_add(self.do_private_notice, target, source, message)
        
    def do_private_notice(self, target, source, message):
        room = self.get_room_for_name(source)
        self.log.debug('Adding notice from %s', room.id)
        room.add_message(message, source, kind='notice')
        self.window.show_all()
    
    async def on_topic_change(self, channel, message, by):
        GLib.idle_add(self.do_topic_change, channel, message, by)
    
    def do_topic_change(self, channel, message, by):
        self.log.debug('Changing topic in %s', channel)
        room = self.get_room_for_name(channel)
        room.topic_pane.update_topic()
        room.add_message(
            f'{by} has changed the channel topic, "{message}"', kind='server'
        )
    
    async def on_mode_change(self, channel, modes, by):
        GLib.idle_add(self.do_mode_change, channel, modes, by)

    def do_mode_change(self, channel, modes, by):
        room = self.get_room_for_name(channel)
        mode_codes = modes[0]
        mode_index = 1
        change = None
        mode_msg = f'{by} '
        try:
            for code in mode_codes:
                if code == '+':
                    change = 'set '
                
                elif code == '-':
                    change = 'removed '
                
                elif code == 'b':
                    mode_msg += f'{change} a ban on {modes[mode_index]}, '
                    mode_index += 1
                
                elif code == 'c':
                    mode_msg += f'{change} colour filter, '
                
                elif code == 'e':
                    mode_msg += f'{change} a ban exemption on {modes[mode_index]}, '
                    mode_index += 1
                
                elif code == 'C':
                    mode_msg += f'{change} CTCP blocking, '
                
                elif code == 'i':
                    mode_msg += f'{change} invite-only, '

                elif code == 'o':
                    mode_msg += f'{change} op on {modes[mode_index]}, '
                    mode_index += 1
                
                elif code == 'v':
                    mode_msg += f'{change} voice on {modes[mode_index]}, '
                    mode_index += 1
                
                elif code == 'q':
                    mode_msg += f'{change} quiet on {modes[mode_index]}, '
                    mode_index += 1
                
                elif code == 'm':
                    mode_msg += f'{change} moderated, '
        
        except IndexError:
            print('No index %s', mode_index)
        
        mode_msg = mode_msg.strip()
        mode_msg = mode_msg.strip(',')
        room.add_message(f'{mode_msg}.', kind='server')

    async def on_user_invite(self, target, channel, by):
        GLib.idle_add(self.do_user_invite, target, channel, by)
    
    def do_user_invite(self, target, channel, by):
        room = self.get_room_for_name(channel)
        room.add_message(f'{by} invited {target}', kind='server')

    async def on_invite(self, channel, by):
        GLib.idle_add(self.do_invite, channel, by)
    
    def do_invite(self, channel, by):
        self.server_room.add_message(
            f'You were invited to {channel} by {by}', kind='server'
        )
        self.server_room.row.set_icon('emblem-important-symbolic')


    async def on_ctcp_action(self, target, source, action):
        self.log.debug('Action in %s from %s: %s %s', target, source, source, action )
        GLib.idle_add(self.do_ctcp_action, target, source, action)
        
    def do_ctcp_action(self, target, source, action):
        self.log.debug('Adding action from %s to %s', source, target)
        if target.startswith('#'):
            room = self.get_room_for_name(target)
        elif target == self.nickname:
            room = self.get_room_for_name(source)
        
        message = f'{source} {action}'
        room.add_message(message, kind='action')
        self.window.show_all()
        room.update_tab_complete(source)

    # Data for this object.
    @property
    def name(self):
        """str: The name of this network (and its room)."""
        return self._config['name']
    
    @name.setter
    def name(self, name):
        """This is actually tracked by the room."""
        self.log.debug('Setting name to %s', name)
        self._config['name'] = name
    
    @property
    def auth(self):
        """str: One of 'sasl', 'pass', or 'none'."""
        return self._config['auth']
    
    @auth.setter
    def auth(self, auth):
        """Only set if it's a valid value."""
        if auth == 'sasl' or auth == 'pass' or auth == 'none':
            self._config['auth'] = auth

    @property
    def host(self):
        """str: The hostname of the server to connect to."""
        return self._config['host']
    
    @host.setter
    def host(self, host):
        self._config['host'] = host
    
    @property
    def port(self):
        return self._config['port']
    
    @port.setter
    def port(self, port):
        """ Only set a port that is within the valid range."""
        if port > 0 and port <= 65535:
            self._config['port'] = int(port)

    @property
    def tls(self):
        """bool: Whether or not to use TLS"""
        return self._config['tls']
    
    @tls.setter
    def tls(self, tls):
        self._config['tls'] = tls

    @property
    def nickname(self):
        """str: The user's nickname"""
        return self._config['nickname']
    
    @nickname.setter
    def nickname(self, nickname):
        self.log.debug('Setting nickname to %s', nickname)
        self._config['nickname'] = nickname

    @property
    def username(self):
        """str: The username to use for the connection"""
        return self._config['username']
    
    @username.setter
    def username(self, username):
        self.log.debug('Setting username to %s', username)
        self._config['username'] = username

    @property
    def realname(self):
        """str: The user's real name"""
        return self._config['realname']
    
    @realname.setter
    def realname(self, realname):
        self._config['realname'] = realname

    @property
    def password(self):
        """str: The user's password."""
        return self._config['password']
    
    @password.setter
    def password(self, password):
        self.log.debug('Setting password')
        self._config['password'] = password



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  ListBoxRows for messages.
"""

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Pango': '1.0'
    }
)
from gi.repository import Gtk, GLib, Pango

from ad1459.formatting import Parser

class MessageRow(Gtk.ListBoxRow):
    """ A ListBoxRow representing a message.

    The attributes for this class are stored directly in the Gtk.Label widgets
    drawn within these rows, as this obviates the need to store extra variables
    in memory.

    Attributes:
        time (str): The time the message was sent.
        sender (str): The user/network who sent the message. "*" by default.
        text (str): The text of the message.
    """

    def __init__(self):
        Gtk.ListBoxRow.__init__(self)
        self.props.selectable = False

        Gtk.StyleContext.add_class(self.get_style_context(), "message-row")
        self.props.margin = 1

        message_grid = Gtk.Grid()
        message_grid.set_hexpand(True)
        message_grid.set_margin_left(12)
        message_grid.set_margin_right(12)
        message_grid.set_column_spacing(12)
        self.add(message_grid)

        self.message_time = Gtk.Label()
        self.message_time.set_selectable(True)
        self.message_time.set_use_markup(True)
        self.message_time.props.halign = Gtk.Align.END
        self.message_time.props.valign = Gtk.Align.START
        self.message_time.props.opacity = 0.5
        message_grid.attach(self.message_time, 0, 1, 1, 1)
        self.message_sender = Gtk.Label()
        self.message_sender.set_selectable(True)
        self.message_sender.set_use_markup(True)
        self.message_sender.props.xalign = 1
        self.message_sender.props.halign = Gtk.Align.END
        self.message_sender.props.valign = Gtk.Align.END
        self.message_sender.props.width_request = 135
        self.message_sender.props.opacity = 0.8
        message_grid.attach(self.message_sender, 0, 0, 1, 1)
        self.message_text = Gtk.Label()
        self.message_text.set_selectable(True)
        self.message_text.set_use_markup(False)
        self.message_text.props.halign = Gtk.Align.START
        self.message_text.set_line_wrap(True)
        self.message_text.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.message_text.set_hexpand(True)
        self.message_text.props.xalign = 0
        message_grid.attach(self.message_text, 2, 0, 1, 2)

        self.message_text.show()
        self.message_time.show()
        self.message_sender.show()

        self.parser = Parser()
        self.kind = 'message'
    
    @property
    def kind(self):
        """str: The type of message this is, based on who sent it and 
        its contents.
        """
        return self._kind
    
    @kind.setter
    def kind(self, kind):
        self._kind = kind
        Gtk.StyleContext.add_class(self.get_style_context(), kind)
        if kind == 'server':
            self.message_text.props.opacity = 0.7

    @property
    def time(self):
        """str: The time the message was sent."""
        return self.message_time.get_text()
    
    @time.setter
    def time(self, time):
        self.message_time.set_markup(f'<i>{time}</i>')
    
    @property
    def sender(self):
        """ The sender of the message, or * for none/network. """
        return self.message_sender.get_text()
    
    @sender.setter
    def sender(self, sender):
        self.message_sender.set_markup(sender)
    
    @property
    def text(self):
        """ The message text."""
        return self.message_text.get_text()
    
    @text.setter
    def text(self, text):
        """ We need to make some conversions depending on the type of message
        this is.
        """
        if self.kind == 'action':
            text = f'\x1D{self.sender} {text}\x1D'
            self.sender = '*'
            
        escaped_text = GLib.markup_escape_text(text, len(text.encode('utf-8')))
        formatted_text = self.parser.parse_text(escaped_text)
        linked_text = self.parser.hyperlinks(formatted_text)

        self.message_text.set_markup(linked_text)
    
    def show_all_contents(self):
        self.show_all()



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This is the Python package.
"""

from setuptools import setup, find_packages, Command
import subprocess

version = {}
with open('ad1459/__version__.py') as fp:
    exec(fp.read(), version)

with open('README.md') as readme:
    long_description = readme.read()

class Release(Command):
    """ Generate a release and push it to git."""
    description = "Generate a release and push it to git."

    user_options = [
        ('dry-run', None, 'Skip the actual release and do a dry run instead.'),
        ('prerelease', None, 'Release this version as a pre-release.'),
        ('force-version=', None, 'Force the version to update to the given value.')
    ]

    def initialize_options(self):
        self.dry_run = False
        self.prerelease = False
        self.force_version = None
    
    def finalize_options(self):
        if self.force_version:
            if not isinstance(self.force_version, str):
                raise Exception('Please specify the test version to release')

    def run(self):
        command = ['npx', 'standard-version@next']
        if self.dry_run:
            command.append('--dry-run')
        if self.prerelease:
            command.append('--prerelease')
        if self.force_version:
            # See https://github.com/conventional-changelog/standard-version#release-as-a-target-type-imperatively-npm-version-like
            command.append('--release-as')
            command.append(self.force_version)
        print(command)
        subprocess.run(command)
        if not self.dry_run:
            subprocess.run(
                ['git', 'push', '--follow-tags']
            )

setup(
    name='ad1459',
    version=version['__version__'],
    packages=find_packages(),
    scripts=['ad1459/ad1459'],

    # Dependencies
    install_requires=[
      'pydle',
      'pure-sasl',
      'keyring'
    ],

    data_files=[
      ('/usr/share/applications', ['data/in.donotspellitgav.ad1459.desktop']),
      ('/usr/share/icons/hicolor/scalable/apps', [
          'data/in.donotspellitgav.Ad1459.svg',
          'data/in.donotspellitgav.Ad1459.Devel.svg'
      ]),
      ('/usr/share/icons/hicolor/symbolic/apps', [
          'data/in.donotspellitgav.Ad1459-symbolic.svg'
      ])
    ],

    # Commands
    cmdclass={'release': Release},

    # Project Metadata
    author='Gaven Royer',
    author_email='gavroyer@gmail.com',
    description='An IRC Client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='ad1459 irc client chat gui gtk',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Environment :: X11 Applications :: Gnome",
        "Environment :: X11 Applications :: GTK",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Topic :: Communications :: Chat :: Internet Relay Chat"
    ]
)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  Then entry for using IRC.
"""

import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

class IrcEntry(Gtk.Entry):
    
    def __init__(self, parent, placeholder='Enter a message'):
        self.log = logging.getLogger('ad1459.ircentry')
        super().__init__()

        self.prematched = False
        self.parent = parent
        self.set_hexpand(True)
        self.set_placeholder_text(placeholder)
        self.props.show_emoji_icon = True
        self.props.max_width_chars = 5000
        self.possible_completions = []

        self.connect('key-press-event', self.on_key_press_event)
    
    def on_key_press_event(self, entry, event):
        if event.keyval == Gdk.keyval_from_name('Tab'):
            # We should currently get the most recent word
            # TODO: Improve this to get the current word at the cursor
            text = self.get_text()
            text_list = text.split()
            
            try:
                current_word = text_list.pop()
            except IndexError:
                return True

            room = self.parent.message_stack.get_visible_child().room
            complete_list = room.users
            for user in room.tab_complete:
                if user in complete_list[::-1]:
                    complete_list.remove(user)
                    complete_list.insert(0, user)

            print(room.tab_complete)
            self.log.debug('Completing word %s', current_word)
            if not self.possible_completions:
                self.log.debug('Users to complete from: %s', complete_list)
                for user in complete_list:
                    if user.lower().startswith(current_word.lower()):
                        self.possible_completions.insert(0, user)
            self.log.debug(
                'Possible completions of %s: %s', 
                current_word,
                self.possible_completions
            )
            
            try:
                completion = self.possible_completions.pop()
                self.possible_completions.insert(0, completion)

                if len(text_list) == 0:
                    completion = f'{completion}:'

                completion = f'{completion} '
                text_list.append(completion)
                text = ' '.join(text_list)
                self.set_text(text)
                length = len(text)
                self.set_position(length)
            except IndexError:
                pass
            self.grab_focus_without_selecting()
            return True
        
        self.possible_completions = []



#!/usr/bin/env python3

""" AD1459, an IRC Client

  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at http://mozilla.org/MPL/2.0/.

  This is the widget module.
"""